from abc import ABC
from threading import RLock
from mq4hemc import HemcQueue, HemcQueueSender, HemcMessage


class HemcObserver(ABC):
    def __init_subclass__(cls):
        super().__init_subclass__()
        cls._ret_dict = {}
        cls._observers = []

    def __init__(self, name: str = "default"):
        if not isinstance(name, str):
            raise ValueError("name must be a string")
        self.name = name
        self._observables = {}
        self._observers.append(self)

    def get_observables_copy(self):
        observables = self._observables.copy()
        return observables

    def observe(self, msg_id, callback):
        if not isinstance(msg_id, str):
            raise ValueError("msg_id must be a string")
        if not callable(callback):
            raise ValueError("callback must be a callable function or method")
        self._observables[msg_id] = callback

    @classmethod
    def fire(cls, msg: HemcMessage, ret_dict: dict = None):
        cls._ret_dict.clear()
        observers = cls._observers.copy()
        for observer in observers:
            observables = observer.get_observables_copy()
            if msg.type in observables:
                callback = observables[msg.type]
                print(f"observer.name: {observer.name}")
                cls._ret_dict[observer.name] = callback(msg)
        if ret_dict is not None:
            ret_dict.update(cls._ret_dict)

    @classmethod
    def clear(cls):
        for observer in cls._observers:
            observer._observables.clear()
        cls._observers.clear()


class HemcObserverEvent(ABC):
    def __init_subclass__(cls, observer_class: HemcObserver):
        super().__init_subclass__()
        if not issubclass(observer_class, HemcObserver):
            raise TypeError("observer_class must be a subclass of HemcObserver")
        cls._observer_class = observer_class

    def __init__(self, msg: HemcMessage, ret_dict: dict = None, autofire: bool = True):
        self.msg = msg
        if autofire:
            self.fire(ret_dict)

    def fire(self, ret_dict: dict = None):
        self._observer_class.fire(self.msg, ret_dict=ret_dict)



from threading import Timer

class HemcTick:
    def __init__(self, interval, callback_tick):
        if not isinstance(interval, (int, float)):
            raise ValueError("interval must be a number")
        self.interval = interval
        if not callable(callback_tick):
            raise ValueError("callback_tick must be a callable function or method")
        self.callback = callback_tick
        self.timer = None

    def start(self):
        self.timer = Timer(self.interval, self.send_tick_event)
        self.timer.daemon = True
        self.timer.start()

    def send_tick_event(self):
        self.callback()
        self.start()



import argparse
import subprocess
import toml

# Create the parser
parser = argparse.ArgumentParser(description='Generate pyproject.toml file.')
parser.add_argument('Path', metavar='path', type=str, help='the path to the pyproject.toml.in file')

# Parse the arguments
args = parser.parse_args()

# Get the latest tag from git
tag = subprocess.getoutput('git describe --abbrev=4 --dirty --always')

# Load the pyproject.toml file
with open(args.Path, 'r') as file:
    pyproject = toml.load(file)

# Update the version
if 'tool' in pyproject and 'poetry' in pyproject['tool'] and 'version' in pyproject['tool']['poetry']:
    pyproject['tool']['poetry']['version'] = tag

with open('pyproject.toml', 'w') as file:
    file.write("# This file is generated by scripts/generate-pyproject.py.\n")
    file.write("# Do not change it by hand, instead modify pyproject.toml.in.\n")
    file.write("# Create a tag using: git tag -a 0.0.10\n")
    file.write("# This tag will be used as a version in the pyproject.toml.\n")

# Write the pyproject.toml file
with open('pyproject.toml', 'a') as file:
    toml.dump(pyproject, file)


import threading
from mq4hemc import HemcQueue, HemcQueueSender, HemcMessage


class HemcService(threading.Thread):
    def __init__(self, process_cb=None):
        # Initialize the message queue
        self._message_queue = HemcQueue(process_item_cb=self._process_item)
        self._running = False
        self._process_cb = None
        if process_cb is not None:
            if not callable(process_cb):
                raise ValueError("process_cb must be a callable function!")
            self._process_cb = process_cb
        threading.Thread.__init__(self)

    def _process_item(self, item: HemcMessage):
        ret = None
        if item.type != "__stop__" and self._process_cb is not None:
            ret = self._process_cb(item)
        return ret

    def _get_sender(self):
        return HemcQueueSender(self._message_queue)

    def start(self) -> None:
        if self._running:
            return
        self._running = True
        return super().start()

    def send_sync_msg(self, message: HemcMessage):
        return self._get_sender().send_wait_reply(message)

    def send_async_msg(self, message: HemcMessage):
        return self._get_sender().send(message)

    def stop(self):
        self._running = False
        message = HemcMessage()
        message.type = "__stop__"
        self.send_async_msg(message)

    def register_process_cb(self, process_cb):
        if not callable(process_cb):
            raise ValueError("process_cb must be a callable function!")
        self._process_cb = process_cb

    def run(self):
        # The main execution loop of the thread.
        while self._running:
            # Get message from the queue and process it
            self._message_queue.get_process()



import threading
import os
import sys
import unittest
from unittest.mock import Mock
import logging
from dataclasses import dataclass, field

# Force insert the path to the beginning of sys.path
# to use the local package instead of the installed package.
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../src')))
from mq4hemc import HemcQueue, HemcQueueSender, HemcMessage

"""
To run this test, run the following commands:
make venv
source ./venv/bin/activate
python3 tests/test_mq4hemc.py

To run all unittests from the root directory, run the following command:
make test

To install the package locally, run the following command:
make install
"""

@dataclass
class BigHemcMessage(HemcMessage):
    payload: dict = field(default_factory=dict)

class TestMq4Hemc(unittest.TestCase):
    def test_send_wait_reply(self):
        mock_process_message = Mock()
        mock_process_message.return_value = "success"

        queue = HemcQueue(process_item_cb = mock_process_message)
        test_thread = threading.Thread(target=queue.get_process, args=(True,))
        test_thread.start()

        sender = HemcQueueSender(queue)
        message = HemcMessage()
        message.type = "test1"
        # the return value of send_wait_reply() is the return value of HemcQueue.process_item_cb()
        status = sender.send_wait_reply(message)
        # Wait for return from queue.get_process()
        test_thread.join()
        mock_process_message.assert_called_once()
        item = mock_process_message.call_args[0][0]
        assert isinstance(item, HemcMessage)
        assert status == "success"
        assert item.callback == sender.get_status_wrapper
        assert item.type == "test1"

    def test_send_wait_reply_big(self):
        mock_process_message = Mock()
        mock_process_message.return_value = "success"

        queue = HemcQueue(process_item_cb = mock_process_message)
        test_thread = threading.Thread(target=queue.get_process, args=(True,))
        test_thread.start()

        sender = HemcQueueSender(queue)
        message = BigHemcMessage()
        message.type = "test1"
        message.payload = {"key": "value"}
        # the return value of send_wait_reply() is the return value of HemcQueue.process_item_cb()
        status = sender.send_wait_reply(message)
        # Wait for return from queue.get_process()
        test_thread.join()
        mock_process_message.assert_called_once()
        item = mock_process_message.call_args[0][0]
        assert isinstance(item, HemcMessage)
        assert status == "success"
        assert item.callback == sender.get_status_wrapper
        assert item.type == "test1"
        assert item.payload == {"key": "value"}

    def test_send(self):
        mock_process_message = Mock()
        mock_process_message.return_value = "success"

        queue = HemcQueue(process_item_cb = mock_process_message)
        test_thread = threading.Thread(target=queue.get_process, args=(True,))
        test_thread.start()

        sender = HemcQueueSender(queue)
        # As we use 'send' method, the callback should not be called.
        sender.get_status_wrapper = Mock()
        message = HemcMessage()
        message.type = "test2"
        sender.send(message)
        # Wait for return from queue.get_process()
        test_thread.join()

        sender.get_status_wrapper.assert_not_called()
        mock_process_message.assert_called_once()
        item = mock_process_message.call_args[0][0]
        assert isinstance(item, HemcMessage)
        assert item.type == "test2"

if __name__ == "__main__":
    logging.basicConfig(level=logging.WARNING)
    logger = logging.getLogger('test_mq4hemc')

    unittest.main()



import threading
import os
import sys
from dataclasses import dataclass, field
import logging
import time
import unittest
from unittest.mock import Mock

# Force insert the path to the beginning of sys.path
# to use the local package instead of the installed package.
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../src')))
from mq4hemc import HemcMessage, HemcService

"""
To run this test, run the following commands:
make venv
source ./venv/bin/activate
python3 tests/test_real_use.py

To run all unittests from the root directory, run the following command:
make test

To install the package locally, run the following command:
make install
"""

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('test_mq4hemc')

@dataclass
class BigHemcMessage(HemcMessage):
    payload: dict = field(default_factory=dict)

class TestHemcService(unittest.TestCase):
    def test_send_wait_reply(self):
        mock_process_message = Mock()
        mock_process_message.return_value = "__success__"

        service = HemcService(mock_process_message)
        service.start()

        for i in range(3):
            message = BigHemcMessage()
            message.type = f"test{i}"
            message.payload = {"key": f"value{i}"}
            logger.info(f"Send {message} and do not wait for reply.")
            status = service.send_async_msg(message)


        message = BigHemcMessage()
        message.type = "test_sync"
        message.payload = {"key": "value"}
        logger.info(f"Now send {message} and wait for reply...")
        status = service.send_sync_msg(message)
        logger.info(f"Message {message} processed, reply: {status}")
        service.stop()
        service.join()
        for call in mock_process_message.call_args_list:
            print(f"Called with args: {call.args}, kwargs: {call.kwargs}")
        assert mock_process_message.call_count == 4
        call_1 = mock_process_message.call_args_list[0]
        assert call_1[0][0].type == "test0"
        assert call_1[0][0].payload == {"key": "value0"}
        call_2 = mock_process_message.call_args_list[1]
        assert call_2[0][0].type == "test1"
        assert call_2[0][0].payload == {"key": "value1"}
        call_3 = mock_process_message.call_args_list[2]
        assert call_3[0][0].type == "test2"
        assert call_3[0][0].payload == {"key": "value2"}
        call_4 = mock_process_message.call_args_list[3]
        assert call_4[0][0].type == "test_sync"
        assert call_4[0][0].payload == {"key": "value"}
        assert status == "__success__"


if __name__ == "__main__":
    unittest.main()
    """
    def process_cb(item: HemcMessage):
        if hasattr(item, 'payload') and item.payload is not None:
            # Simulate processing time
            time.sleep(1)
        logger.info(f"Processed message '{item.type}', payload: {item.payload}")
        return item.type

    service = HemcService(process_cb)
    service.start()

    for i in range(3):
        message = BigHemcMessage()
        message.type = f"test{i}"
        message.payload = {"key": f"value{i}"}
        logger.info(f"Send {message} and do not wait for reply.")
        status = service.send_async_msg(message)

    message = BigHemcMessage()
    message.type = "test_sync"
    message.payload = {"key": "value"}
    logger.info(f"Now send {message} and wait for reply...")
    status = service.send_sync_msg(message)
    logger.info(f"Message {message} processed, reply: {status}")
    service.stop()
    service.join()
    logger.info("Service stopped.")
    """



import os
import sys
import unittest
from unittest.mock import Mock

# Force insert the path to the beginning of sys.path
# to use the local package instead of the installed package.
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../src')))
from mq4hemc import HemcObserver, HemcObserverEvent, HemcMessage

"""
To run this test, run the following commands:
make venv
source ./venv/bin/activate
python3 tests/test_observer.py

To run all unittests from the root directory, run the following command:
make test

To install the package locally, run the following command:
make install
"""


class GNSSObserver(HemcObserver):
    pass


class GNSSObserverEvent(HemcObserverEvent, observer_class=GNSSObserver):
    pass


class AnotherObserver(HemcObserver):
    pass


class AnotherObserverEvent(HemcObserverEvent, observer_class=AnotherObserver):
    pass


class TestObserverEventM(unittest.TestCase):
    def test_init(self):
        GNSSObserver.clear()

        mock_process_message = Mock()
        mock_process_message.return_value = "success"

        gnss_observer = GNSSObserver()
        gnss_observer.observe("test_id", mock_process_message)

        GNSSObserverEvent(HemcMessage(type="test_id"))
        # assert mock_process_message.called_once_with(msg_id, msg_data)
        print(f"gnss_observer._observers: {gnss_observer._observers}")
        mock_process_message1 = Mock()
        mock_process_message1.return_value = "success"

        another_observer = GNSSObserver()
        another_observer.observe("test_id", mock_process_message1)

        GNSSObserverEvent(HemcMessage(type="test_id"))
        assert 2 == mock_process_message.call_count
        for call in mock_process_message.call_args_list:
            args, kwargs = call
            print(f"Args: {args}, Kwargs: {kwargs}")
            assert args == (HemcMessage(type='test_id', callback=None),)
            # print(f"Args: {args}, Kwargs: {kwargs}")
        assert 1 == mock_process_message1.call_count
        assert mock_process_message1.called_once_with((HemcMessage(type='test_id', callback=None),))

        print(f"gnss_observer._observers: {gnss_observer._observers}")
        print(f"another_observer._observers: {another_observer._observers}")

    def test_ret(self):
        GNSSObserver.clear()

        mock_process_message = Mock()
        mock_process_message.return_value = "success"

        gnss_observer = GNSSObserver(name="gnss_observer")
        gnss_observer.observe("test_id", mock_process_message)

        ret_dict = {}
        GNSSObserverEvent(HemcMessage(type="test_id"), ret_dict=ret_dict)
        assert ret_dict == {"gnss_observer": "success"}
        # assert mock_process_message.called_once_with(msg_id, msg_data)
        mock_process_message1 = Mock()
        mock_process_message1.return_value = "success1"
        another_observer = GNSSObserver(name="gnss_observer1")
        another_observer.observe("test_id", mock_process_message1)

        ret_dict.clear()
        GNSSObserverEvent(HemcMessage(type="test_id"), ret_dict=ret_dict)
        print(f"ret_dict: {ret_dict}")
        assert ret_dict == {"gnss_observer": "success", "gnss_observer1": "success1"}
        assert 2 == mock_process_message.call_count
        for call in mock_process_message.call_args_list:
            args, kwargs = call
            print(f"Args: {args}, Kwargs: {kwargs}")
            assert args == (HemcMessage(type='test_id', callback=None),)
        assert mock_process_message1.called_once_with((HemcMessage(type='test_id', callback=None),))
        # observer_class = MagicMock()
        # observer_class._mutex = MagicMock()
        # observer_class._observers = []

        # event = ObserverEventM(msg_id, msg_data, autofire=False)
        # self.assertEqual(event.msg_id, msg_id)
        # self.assertEqual(event.msg_data, msg_data)

    def test_fire(self):
        import copy

        original = [[1, 2, 3], [4, 5, 6]]
        shallow_copy = copy.copy(original)

        original.append([7, 8, 9])
        original[0][0] = 99

        print(shallow_copy)

    def test_fire1(self):
        import copy

        original = [[1, 2, 3], [4, 5, 6]]
        shallow_copy = copy.copy(original)

        shallow_copy[0][0] = 99

        print(original)

    def test_fire2(self):
        def deep_copy_ret(data_dict):
            local_dict = {"key": "value"}
            data_dict.update(local_dict)

        original = {}
        deep_copy_ret(original)

        print(f"return after deepcopy {original}")


if __name__ == "__main__":
    unittest.main()
